/*
 *   COPYRIGHT NOTICE: This code was written by Markus Mummert. All
 *   rights reserved. FOR LICENCE TERMS, SEE ACCOMPANYING FILE "LICENCE".
 */
/************************************************************************/
/*									*/
/*  Datei    :	DFT.C (main)						*/
/*  Version  :	(see below)						*/
/*  Autor    :	Markus Mummert						*/
/*									*/
/*  Kommentar:	sliding discrete Fourier-transform			*/
/*		of signals using FFT					*/
/*									*/
/*  history  :	1999/07/17 begin based on FASTCONV.C/V2.8/1999/02/25	*/
/*		(continued below)					*/
/*									*/
/************************************************************************/
/*
 *  $Source: /home/mummert/src/MUMTOOLS/dft/RCS/dft.c,v $
 *  $Revision: 1.3 $
 *  $Date: 1999/11/14 20:23:24 $
 *  $Author: mummert $
 *  $Log: dft.c,v $
 *  Revision 1.3  1999/11/14 20:23:24  mummert
 *  OENDIDX is last index instead of OENDIDX-1 (NOT BACKWARD COMPATIBLE)
 *  FIXANALYTIC and IDELAY implemented; SCALE for EXTWFILE functional;
 *  range of L fits to range of N
 *
 *  Revision 1.2  1999/08/25 18:18:46  mummert
 *  EXTWFILE/EXTWFMT implemented; WINDOW help ouput on stdout
 *
 *  Revision 1.1  1999/08/24 23:36:44  mummert
 *  Multiple window separator now asterisk; WINDOW list formatting changed
 *
 *  Revision 1.0  1999/08/22 22:40:26  mummert
 *  First release
 *
 */
	
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "portab.h"			/* portability			*/
#include "key.h"			/* keytab-management		*/
#include "window.h"
#include "winhelp.h"
#include "fft.h"
#include "numio.h"

int	DEBUG = 0;

#define	DFLTPARMFILE	"dft.par"	/* default parameter file	*/

/*----------------------------------------------------------------------*/
/*	global variables						*/
/*----------------------------------------------------------------------*/

char 	g_ifile[MAXLINE],		/* input file name		*/
	g_ifields[MAXLINE],		/* input fields			*/
	g_ofile[MAXLINE],		/* output file name		*/
	g_wfile[MAXLINE],		/* window file name		*/
	g_ifmt[MAXLINE],		/* input format			*/
	g_ofmt[MAXLINE],		/* output format		*/
	g_wfmt[MAXLINE],		/* window format		*/
	g_ofldfmt[MAXLINE],		/* output-printf()-format	*/
	g_obsep[MAXLINE],		/* output block separator	*/
	g_ofsep[MAXLINE],		/* output field spearator	*/
	g_ofields[MAXLINE],		/* output fields		*/
	g_wtype[MAXLINE],		/* window type			*/
	g_wsmpl[MAXLINE],		/* window sampling		*/
	g_transform[MAXLINE],		/* transform type		*/
	g_sn[MAXLINE],			/* block size N (string)	*/
	g_fixa[MAXLINE],		/* fix level of analytic output	*/
	g_debug[MAXLINE],		/* debug-switch			*/
	g_splfile[MAXLINE],		/* ftt-format level file name	*/
	g_sppfile[MAXLINE];		/* ftp-format phase file name	*/

double  g_scale_usr,			/* user-gain			*/
	g_scale,			/* internal gain		*/
	g_f3db,				/* 3dB-cutoff-f some windows.	*/
	g_fco,				/* cutoff-f for sinc window	*/
	g_walpha,			/* trade-off-parm. (Harris78)	*/
	g_wraise,			/* window-raise			*/
	g_fa;				/* sampling frequency		*/

int	g_n,				/* block size N			*/
	g_rot = 0,			/* block-rotate samples		*/
	g_m,				/* block advance M		*/
	g_idelay,			/* input delay in samples	*/
	g_ofibeg,			/* output from f-idx (modulo n)	*/
	g_ofihop,			/* output idx distance (mod n)	*/
	g_ofiend,			/* output stop f-idx (modulo n)	*/
	g_tshift,			/* time-shift for causal-IIR	*/
	g_nwdow;			/* window size L		*/

/*- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
					/* Keytab (vergl. key.h/c)	*/
KEY_TAB kt[]= {
"IFILE",	g_ifile,	STRING, 
				"Input file (stdin: '-')",
				"-",	EMPTY,
"IFMT",		g_ifmt,		STRING, 
				"Input file format",
				"word",	
"b u w l d f a byte ubyte word long double float ascii",
"IFIELDS",	g_ifields,	STRING, 
				"Input fields (r, i, ri, ir)",
				"r",	"r i ri ir",
"IDELAY",	pt(g_idelay),	INT, 
				"Input delay by zero-samples (-1: =L-1)",
				"0",	"-1 2000000000",
"FSAMP", 	pt(g_fa),	DOUBLE,	
				"Sample-f (for WF3DB, WFCUT, OFIELDS=f)",
				"1.0",
				"1E-50 1E+50",	
"SCALE",	pt(g_scale_usr),DOUBLE,	
				"Gain",
				"1.0", 
				EMPTY,	
"N",		g_sn,		STRING,	
				"Block size (powers of 2 only)",
				"1024",	
"2 4 8 16 32 64 128 256 512 1024 2048 4096 8192 16384 32768 65536\
 131072 262144 524288 1048576",
"L",		pt(g_nwdow),	INT,	
				"Window length (0: =N); zero-pad N-L",
				"0",	"0 1048576",	
"WINDOW",	g_wtype,	STRING,	
				"Window type (see 'WINDOW help')",
				"rectangle",
				EMPTY,
"WSAMPLING",	g_wsmpl,	STRING,
				"auto: (N>L:symmetric, N=L:periodic)",
				"auto",	"auto symmetric periodic a s p",
"WALPHA",	pt(g_walpha),	DOUBLE,	
				"Window parameter as defined by Harris78",
				"1.0",
				EMPTY,	
"WFCUT",	pt(g_fco),	DOUBLE,	
				"Cutoff-frequency for window 'sinc'",
				"0.25",	EMPTY,
"WF3DB",	pt(g_f3db),	DOUBLE,	
				"f3dB for 'gaussian-f3db'/causal-IIR-wdows",
				"0.002",	EMPTY,
"WSHIFT",	pt(g_tshift),	INT,	
				"Sample-shift t=0 of causal-IIR-windows",
				"0",	EMPTY,
"WRAISE",	pt(g_wraise),	DOUBLE,	
				"Raise wdow by WRAISE+(1-WRAISE)*w[i]/wmax",
				"0.0",	EMPTY,
"EXTWFILE",	g_wfile,	STRING, 
				"External window file >= L ('-': internal)",
				"-",	EMPTY,
"EXTWFMT",	g_wfmt,		STRING, 
				"External window file-format",
				"ascii",	
"b u w l d f a byte ubyte word long double float ascii",
/*
 * "ROT",	pt(g_rot),	INT,
 * 				"Rotate block ROT samples",
 *				"0",	EMPTY,
 */
"TRANSFORM",	g_transform,	STRING,
				"Transform type (dft, idft, none)",
				"dft",	"dft idft none",
"FIXANALYTIC",	g_fixa,		STRING,
				"Multiply spectrum by 2 except at i=0,N/2",
				"no",	"auto yes no a y n ",
"M",		pt(g_m),	INT, 
				"Block advance (0: single spectrum)",
				"0",	"0 2000000000",
"OBEGIDX",	pt(g_ofibeg),	INT,
				"Output start index",
				"0",	"0 2000000000",
"OHOPIDX",	pt(g_ofihop),	INT,
				"Output index distance",
				"1",	"0 2000000000",
"OENDIDX",	pt(g_ofiend),	INT,
				"Output last index (-1: =N-1, -2: =N/2)",
				"-2",	"-2 2000000000",
"OFILE",	g_ofile,	STRING,
				"Output file (stdout: '-' pipe: '|...')",
				"-",
				EMPTY,
"OFMT",		g_ofmt,		STRING,	
				"OFILE: file format",
				"ascii",	
"b u w l d f a byte ubyte word long double float ascii",
"OFIELDS",	g_ofields,	STRING,	
				"OFILE: data fields (combine f,r,i,l,p)",
				"lp",
				EMPTY,
"OFIELDFMT",	g_ofldfmt,	STRING,	
				"OFMT=ascii: printf()-format data fields",
				"%+le",
				EMPTY,
"OFIELDSEP",	g_ofsep,	STRING,	
				"OFMT=ascii: data field separator",
				"\\t",
				EMPTY,
"OBLOCKSEP",	g_obsep,	STRING,	
				"OFMT=ascii: block separator",
				"\\n",	EMPTY,
"SPLFILE",	g_splfile,	STRING,
				"ftt-level file (turns OFILE off)",
				"-",	EMPTY,
"SPPFILE",	g_sppfile,	STRING,
				"ftp-phase file (turns OFILE off)",
				"-",	EMPTY,
/*
 * "DEBUG",	g_debug,	STRING,	
 * 				"Debug output",
 *				"no",	"y n yes no",
 */
0L,0L,0L,0L,0L,0L };

/*----------------------------------------------------------------------*/
/*  Write header of usage message, to be called by key.c		*/
/*----------------------------------------------------------------------*/
header()
{
char string[90], version[90];

sscanf("$Revision: 1.3 $ $Date: 1999/11/14 20:23:24 $ $Author: mummert $",
        "$%*s %s $ $%*s %s %*s $ $%*s %s $",
        string, string+30, string+60);
sprintf(version, "DFT V%s %s %s", string, string+30, string+60);
fprintf(stderr, "\n");
fprintf(stderr, "    %s\n\n", version);
fprintf(stderr, "    Sliding Discrete Fourier-Transform of signals");
fprintf(stderr, " using FFT\n");
fprintf(stderr, "\n");
}
/*----------------------------------------------------------------------*/
/*  Calculate level in dB, clip very small levels to -999.999dB		*/
/*----------------------------------------------------------------------*/
double level(re, im)
double re, im;
{
	double pow, mag;
	static double flo=1.0000115136E-50;	/* -9.999999+E2 dB	*/

	pow = re*re + im*im;
	mag = sqrt(pow);

	return(mag < flo ?  20.*log(flo)/M_LN10 : 20.*log(mag)/M_LN10);
}
/*----------------------------------------------------------------------*/
/*  Calculate phase, range [0..2PI] (both interval limits included)	*/
/*----------------------------------------------------------------------*/
double phase(re, im)
double re, im;
{
        double ph, pi = M_PI;

        if (fabs(re) > 1E-64 && fabs(im/re) < 1E+12)
                ph = atan(im/re);
        else {
                if (im > 0)
                        ph = pi/2;
                else if (im < 0)
                        ph = -pi/2;
                else
                        ph = 0;
        }                               /* -> range of ph: -90...90DEG  */
        if (re < 0)
                ph += pi;               /* -> range of ph: -90...270DEG */
        if (ph < 0)
                ph += 2*pi;             /* -> range of ph:   0...360DEG */
        return(ph);
}
/*----------------------------------------------------------------------*/
/*  Insert possible delay before calling numread()			*/
/*  Return number of items read (including delayed)			*/
/*----------------------------------------------------------------------*/
delayed_read(istream, in, delayp, read)
FILE *istream;
double *in;
int *delayp, read;
{
	int i, delay;

	delay = *delayp;
	if (delay >= read) {
	    for (i = 0; i < read; i++)
		in[i] = 0;
	    *delayp -= read;
	    return (read);
	}
	if (delay > 0 /* && delay < read */) {
	    for (i = 0; i < delay; i++)
		in[i] = 0;
	    *delayp = 0;
	}
	return (delay + numread(istream, g_ifmt, in + delay, read - delay));
}
/*----------------------------------------------------------------------*/
/*  Set up <in> to contain next block by shifting/reading data		*/
/*  Return non-zero upon EOF						*/
/*----------------------------------------------------------------------*/
next_block(istream, in)
FILE *istream;
double *in;
{
	int i, skip, read, save;
	static int first_call = 1, have_2_fields, delay;

	if (first_call) {
	    first_call = 0;
	    have_2_fields = (g_ifields[1] ? 1 : 0);
	    delay = g_idelay << have_2_fields;
	    read = g_nwdow << have_2_fields;
	    if (delayed_read(istream, in, &delay, read) != read)
		return(-1);
	    return (0);
	}
	if ((skip = g_m - g_nwdow) > 0) {
		read = g_n << have_2_fields;	/* skip by largest step */
	        for (i = skip/g_n; i > 0; i--)	/* within size of in	*/
		    if (delayed_read(istream, in, &delay, read) != read)
			return(-1);
		read = (skip%g_n) << have_2_fields;
		delayed_read(istream, in, &delay, read);
		read = g_nwdow << have_2_fields;
	        if (delayed_read(istream, in, &delay, read) != read)
		    return(-1); 
	} else {
		save = (g_nwdow - g_m) << have_2_fields;
		read = g_m << have_2_fields;
		for (i = 0; i < save; i++)
		    in[i] = in[i+read];
		if (delayed_read(istream, 
			in + save, &delay, read) != read)
		    return(-1);
	}
	return(0);
}
/*----------------------------------------------------------------------*/
/*  Handle output via OFILE						*/
/*----------------------------------------------------------------------*/
output(xr, xi, out, nofields, ostream)
double *xr, *xi, *out;
int nofields;
FILE *ostream;
{
	int i, k, l, check;
	char *errmsg, s[MAXLINE];

	l = 0;				/* set up output data		*/
	i = g_ofibeg;
	while (-1) {
	    for (k = 0; k < nofields; k++) {
		switch (g_ofields[k] | 0x20) {
		case 'f': out[l+k] = i*g_fa/g_n; break;
		case 'r': out[l+k] = xr[i]; break;
		case 'i': out[l+k] = xi[i]; break;
		case 'l': out[l+k] = level(xr[i], xi[i]);
			  break;
		case 'p': out[l+k] = phase(xr[i], xi[i]);
			  break;
		default:
		    sprintf(s, "unknown field specifier `%c' in OFIELDS",
			    g_ofields[k]);
		    errorhandler(s);
		    exit(-1);
		}
	    }
	    l += nofields;
	    if (i == g_ofiend)
		break;
	    i = (i+g_ofihop+g_n)%g_n;
	}
					/* write data to ostream	*/
	check = numwrite(ostream, g_ofmt, out, l, 
		nofields, g_ofldfmt, g_ofsep);
	if (check == -l) {
		errorhandler("overflow when converting to OFMT" );
		exit(-1);
	} else if (check != l)  {
		errmsg = catmsg(g_ofile, "when writing");
		errorhandler(errmsg);
		exit(-1);
	}
	if ((g_ofmt[0] | 0x20) == 'a')
		fprintf(ostream, g_obsep);
}
/*----------------------------------------------------------------------*/
/*  Handle level-output via SPLFILE					*/
/*----------------------------------------------------------------------*/
output_splstream(xr, xi, out, splstream)
double *xr, *xi, *out;
FILE *splstream;
{
	int i, l, check;
	double d;
	char *errmsg;

	l = 0;				/* set up output data		*/
	i = g_ofibeg;
	while (-1) {
		/* ftt-format: 0dB:180, 0:<=-90dB */  
		d = (180.0 + 2*level(xr[i], xi[i]))/256.0;
		out[l] = (d < 0.0 ? 0.0 : d);
		l++;
		if (i == g_ofiend)
			break;
		i = (i+g_ofihop+g_n)%g_n;
	}
					/* write data to splstream	*/
	check = numwrite(splstream, "u", out, l, 0, NULL, NULL);
	if (check == -l) {
		errorhandler("overflow when converting to SPLFILE-format" );
		exit(-1);
	} else if (check != l)  {
		errmsg = catmsg(g_splfile, "when writing");
		errorhandler(errmsg);
		exit(-1);
	}
}
/*----------------------------------------------------------------------*/
/*  Handle level-output via SPPFILE					*/
/*----------------------------------------------------------------------*/
output_sppstream(xr, xi, out, sppstream)
double *xr, *xi, *out;
FILE *sppstream;
{
	int i, l, check;
	char *errmsg;

	l = 0;				/* set up output data		*/
	i = g_ofibeg;
	while (-1) {
		/* decrease phase-range by NUMIO_DELTA to avoid upper	*/
		/* limit (see phase())					*/
		out[l] = phase(xr[i], xi[i]) * (1-NUMIO_DELTA) / (2*M_PI);
		l++;
		if (i == g_ofiend)
			break;
		i = (i+g_ofihop+g_n)%g_n;
	}
					/* write data to sppstream	*/
	check = numwrite(sppstream, "u", out, l, 0, NULL, NULL);
	if (check == -l) {
		errorhandler("overflow when converting to SPPFILE-format" );
		exit(-1);
	} else if (check != l)  {
		errmsg = catmsg(g_sppfile, "when writing");
		errorhandler(errmsg);
		exit(-1);
	}
}
/*----------------------------------------------------------------------*/
/*  load window								*/
/*----------------------------------------------------------------------*/
void load_window(wstream, wdow)
FILE *wstream;
double *wdow;
{
	int i;
	char *errmsg;
	double wnorm;

	if (numread(wstream, g_wfmt, wdow, g_nwdow) != g_nwdow) {
		errmsg = catmsg(g_wfile, "not enough (<L) samples found");
		errorhandler(errmsg);
		exit(-1);
	}
	wnorm = 0.0;			/* normalize			*/
	for (i = 0; i < g_nwdow; i++)
		wnorm += wdow[i];
	if (fabs (wnorm) < 1E-30) {
		errmsg = catmsg(g_wfile, "has zero gain (arithmetic mean)");
		errorhandler(errmsg);
		exit(-1);
	}
	wnorm /= g_scale;
	for (i = 0; i < g_nwdow; i++)
		wdow[i] /= wnorm;	    
}
/*----------------------------------------------------------------------*/
/*  Make window								*/
/*----------------------------------------------------------------------*/
void make_window(wdow)
double *wdow;
{
	int i, wsampling;
	char *p, w[MAXLINE], em[MAXLINE];


	wsampling = WINDOW_SYMMETRIC; 
	switch (g_wsmpl[0] | 0x20) {
	case 'a': if (g_nwdow == g_n)
			wsampling = WINDOW_PERIODIC;
		  break;
	case 'p': wsampling = WINDOW_PERIODIC; 
		  break;
	case 's': break;
	}

	p = g_wtype; i = 0;
	sscanf(p += i, "%[0-9a-zA-Z-]%n", w, &i);
	do {
	    if(!strcmp(w, "list")) {
		list_of_windows(stdout);
		exit(-1);
	    } else if(!strcmp(w, "help")) {
		help_on_windows(stdout);
		exit(-1);
	    } else if(!strcmp(w, "notes")) {
		notes_on_windows(stdout);
		exit(-1);
	    }
	    switch (window_window(wdow, g_nwdow, w, g_scale, wsampling,
		    g_walpha, g_f3db/g_fa, g_fco/g_fa, g_tshift)) {
	    case WINDOW_OK:
		break;
	    case WINDOW_ERROR_NOMEM:
		errorhandler("not enough memory");
		exit(-1);
	    case WINDOW_ERROR_ZERODC:
		errorhandler("resulting window is zero due to TSHIFT");
		exit(-1);
	    case WINDOW_ERROR_PRANGE:
		sprintf(em, 
		    "WINDOW parameter out of range for `%s'", w);
		errorhandler(em);
		exit(-1);
	    case WINDOW_ERROR_NUMERIC:
		sprintf(em, 
		    "numerical error calculating window `%s'", w);
		errorhandler(em);
		exit(-1);
	    case WINDOW_ERROR_NOWIN:
		sprintf(em, 
		    "unknown type `%s' in WINDOW, try `WINDOW help'", w);
		errorhandler(em);
		exit(-1);
	    default:
		errorhandler("UNEXPECTED: window_window(): unknown error");
		exit(-1);
	    }
	} while (sscanf(p += i, "%[*]%[0-9a-zA-Z-]%n", w, w, &i) >= 2);

	if (fabs(g_wraise) < 1E-30)
	        return;
	switch (window_raise(wdow, g_nwdow, g_wraise, g_scale)) {
	case WINDOW_OK:
		break;
	case WINDOW_ERROR_ZERODC:
		errorhandler("resulting window has zero gain due to RAISE");
		exit(-1);
	default:
		errorhandler("UNEXPECTED: window_raise(): unknown error");
		exit(-1);
	}
}	
/*----------------------------------------------------------------------*/
/*  Central processing loop						*/
/*----------------------------------------------------------------------*/
do_calc(in, istream, wstream, wdow, fftp, fixflag, out, nofields, ostream, 
	splstream, sppstream)
double *in, *wdow, *out;
FFTDATA *fftp;
FILE *istream, *ostream, *wstream, *splstream, *sppstream;
int fixflag, nofields;
{
	int i, k;
	char *errmsg;
	FFTELEM *xr, *xi;

	if (wstream != NULL)
	    load_window(wstream, wdow);	/* load	external window		*/
	else
	    make_window(wdow);		/* calculate internal window	*/

					/* read samples, rotated inser-	*/
					/* tion into fft-array		*/
	xr = fftp->r;
	xi = fftp->i;
	if (next_block(istream, in)) {
		errmsg = catmsg(g_ifile, "not enough (<L) samples found");
		errorhandler(errmsg);
		exit(-1);
	};

	while(-1) {
	    for (i = 0; i < g_nwdow; i++) {
		k = (i+g_rot+g_n)%g_n;
		if (g_ifields[0] == 'r') {
		  if (g_ifields[1]) { /* ri */
		    xr[k] = (FFTELEM)(wdow[i] * in[i<<1]);
		    xi[k] = (FFTELEM)(wdow[i] * in[(i<<1)+1]);
		  } else { /* r */
		    xr[k] = (FFTELEM)(wdow[i] * in[i]);
		    xi[k] = 0;
		  }
		} else {
		  if (g_ifields[1]) { /* ir */
		    xr[k] = (FFTELEM)(wdow[i] * in[(i<<1)+1]);
		    xi[k] = (FFTELEM)(wdow[i] * in[i<<1]);
		  } else { /* i */
		    xr[k] = 0;
		    xi[k] = (FFTELEM)(wdow[i] * in[i]);
		  }
		}
	    }					/* zero-padding		*/
	    for (i = g_nwdow; i < g_n; i++) {
		k = (i+g_rot+g_n)%g_n;
		xr[k] = 0;
		xi[k] = 0;
	    } 

	    if (g_transform[0] == 'i')
		ifft(fftp);			/* inverse FFT, or	*/
	    else if (g_transform[0] == 'd')
		fft(fftp);			/* FFT, or nothing	*/

	    if (fixflag) {			/* fix analytic	spectr.	*/
		xr[0] /= 2;
		xi[0] /= 2;
		xr[g_n/2] /= 2;
		xi[g_n/2] /= 2;
	    }

	    if (ostream != NULL)
		output(xr, xi, out, nofields, ostream);
	    if (splstream != NULL)
		output_splstream(xr, xi, out, splstream);
	    if (sppstream != NULL)
		output_sppstream(xr, xi, out, sppstream);

	    if (g_m <= 0)
		return; 
	    if (next_block(istream, in))
		return;
	}
}
/*----------------------------------------------------------------------*/
/*  main								*/
/*----------------------------------------------------------------------*/
main(argc, argv)
int argc;
char *argv[];
{
	char *errmsg, c;
	double *in, *out, *wdow;
	FILE *istream, *ostream, *wstream, *splstream, *sppstream;
	FFTDATA *fftp;
	int d, k, i, nofields, fftsize, ldn, check, fixflag;

					/* keytab-mechanism		*/
	if (initkeytab(argc, argv, kt, DFLTPARMFILE))
		exit(-1);
	if ((g_debug[0] | 0x20) != 'n')
		DEBUG = 1;

					/* post-process keytab values..	*/
	sscanf(g_sn, "%ld", &g_n);
	ldn = -1;
	while ((g_n >> 1) >> ++ldn)
	  ;
	fftsize = 1 << ldn; 
	if (fftsize != g_n) {
 		errorhandler("UNEXPECTED: fftsize != g_n");
		exit(-1);
	}
	if ((nofields = strlen(g_ofields)) > 5) {
		errorhandler("too many fields in OFIELDS");
		exit(-1);
	}
	if (g_nwdow >  g_n) {
		errorhandler("L > N not allowed");
		exit(-1);
	}
	if (g_nwdow <= 0)
		g_nwdow = g_n;
	if (g_idelay == -1)
		g_idelay = g_nwdow - 1;
	if (g_ofiend == -2)		/* case of -1 handled by modulo	*/
		g_ofiend = g_n/2;

	g_rot=((g_rot%g_n)+g_n)%g_n;	/* enforce nonnegative modulo N	*/

	g_ofibeg = ((g_ofibeg % g_n) + g_n) % g_n;
	g_ofiend = ((g_ofiend % g_n) + g_n) % g_n;
	g_ofihop = ((g_ofihop % g_n) + g_n) % g_n;
	if (g_ofihop == 0) {
		errorhandler("illegal OHOPIDX");
		exit(-1);
	}
	/* 
	 * Tricky: trim end to be ((k-1)*hop+beg)%N == end with 0<k<=N.
	 * This is important for loop termination-condition somewhere.
	 * d is the number of indicies to be output if hop=1.
	 * k is the number of indicies to be output for given hop.
	 * hop, end, beg must be of non-negative modulo N, also hop > 0;
	 */
	d = ((g_ofiend - g_ofibeg + g_n) % g_n) + 1;
        k = (d + (g_ofihop - 1)) / g_ofihop;
	g_ofiend = ((k-1) * g_ofihop + g_ofibeg);
	g_ofiend = ((g_ofiend % g_n) + g_n) % g_n;

	fixflag = 0;			/* handle FIXANALYTIC		*/
	switch (g_fixa[0] | 0x20) {
	case 'a':
		if (g_ofibeg <= g_n/2)
		  if (g_ofibeg + (k-1)*g_ofihop <= g_n/2) 
		    fixflag = 1;
		if (g_ofibeg >= g_n/2)
		  if (g_ofibeg + (k-1)*g_ofihop <= g_n) 
		    fixflag = 1;
		break;
	case 'y':
		fixflag = 1; break;
	default:
		break;
	}

	g_scale = g_scale_usr;
	if (fixflag)			/* FIXANALYTIC: mply. all by 2	*/
		g_scale *= 2;		/* div. f-idx 0,N/2 by 2 later	*/

	c = g_ofmt[0] | 0x20;		/* check for illegal fields	*/
	if (g_ofile[0] != 0 && c != 'a' && c != 'f' && c != 'd') {
	    for (i = 0; i < nofields; i++) {
		c = g_ofields[i];
		if (c == 'f' || c == 'l' || c == 'p') {
	    errorhandler("OFIELDS f,l,p not allowed with OFMT=u,b,w,l");
	    exit(-1);
		}
	    }
	}
					/* convert `\\'-sequences	*/
	strcpy(g_wtype, convspchar(g_wtype));
	strcpy(g_ofsep, convspchar(g_ofsep));
	strcpy(g_obsep, convspchar(g_obsep));
	strcpy(g_ofmt, convspchar(g_ofmt));
					/* allocate fftsize for complex */
	if ((fftp = fft_init(ldn))				== NULL ||
	    (wdow = window_init(fftsize*2))			== NULL ||
	    (in  = (double*)(malloc(sizeof(double)*fftsize*2)))	== NULL ||
	    (out = (double*)(malloc(sizeof(double)*fftsize*nofields)))
								== NULL) {
		errorhandler("not enough memory");
		exit(-1);
	}
					/* open input file		*/
	istream = stdin;
	if (g_ifile[0] != '-' || g_ifile[1] != 0) {
		if ((istream = fopen(g_ifile, READ)) == NULL) {
			errmsg = catmsg(g_ifile, "cannot open");
			errorhandler(errmsg);
			exit(-1);
		}
	}
					/* open window file		*/
	wstream = NULL;
	if (g_wfile[0] != '-' || g_wfile[1] != 0) {
		if ((wstream = fopen(g_wfile, READ)) == NULL) {
			errmsg = catmsg(g_wfile, "cannot open");
			errorhandler(errmsg);
			exit(-1);
		}
	}
					/* open output ftt/ftp-file	*/
	splstream = NULL;
	if (g_splfile[0] != '-' || g_splfile[1] != 0) {
		if ((splstream = fopen(g_splfile, WRITE)) == NULL) {
			errmsg = catmsg(g_splfile, "cannot open");
			errorhandler(errmsg);
			exit(-1);
		}
	}
	sppstream = NULL;
	if (g_sppfile[0] != '-' || g_sppfile[1] != 0) {
		if ((sppstream = fopen(g_sppfile, WRITE)) == NULL) {
			errmsg = catmsg(g_sppfile, "cannot open");
			errorhandler(errmsg);
			exit(-1);
		}
	}
					/* open output file/pipe	*/
	ostream = NULL;			/* only if no ftp- or ftt-file	*/
	if (splstream == NULL && sppstream == NULL) {
	  if (g_ofile[0] != '-' || g_ofile[1] != 0) {
	    if(g_ofile[0] == '|') {
		if ((ostream = popen(g_ofile+1, WRITE)) == NULL) {
			errmsg = catmsg(g_ofile+1, "cannot execute");
			errorhandler(errmsg);
			exit(-1);
		}
	    } else {
		if ((ostream = fopen(g_ofile, WRITE)) == NULL) {
			errmsg = catmsg(g_ofile, "cannot open");
			errorhandler(errmsg);
			exit(-1);
		}
	    }
	  } else {
		ostream = stdout;
	  }
	}
					/* do the actual processing	*/
	do_calc(in, istream, wstream, wdow, fftp, fixflag, out, nofields, 
		ostream, splstream, sppstream);

					/* clean-up ...			*/
	fclose(istream);
	if (ostream != NULL) {
	    if (g_ofile[0] == '|')
	        pclose(ostream);
	    else
	        fclose(ostream);
	}
	if (wstream != NULL)
		fclose(wstream);
	if (splstream != NULL)
		fclose(ostream);
	if (splstream != NULL)
		fclose(ostream);

	free(out);
	free(in);
	window_discard(wdow);
	fft_discard(fftp);
	exit(0);
}
/*******EOT**************************************************************/
